using System;
using iNTERFACEWARE.Chameleon;
using System.Reflection;

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
// NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// iNTERFACEWARE BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CON-
// NECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

namespace Hl7Service
{
    class ListenerModule
    {
        private bool runningAsService;
        private LlpServer listener;
        private Engine autoAckEngine;
        private System.IO.StreamWriter logStream;

        private static string APP_AUTOACK_VMD_FILE = "autoack.vmd";
        private static string APP_LOG_FILE = "hl7log.txt";
        private static ushort APP_LISTEN_PORT = 5123;
        private static int APP_MAX_MESSAGE_LOG_LENGTH = 2048;

        public ListenerModule(bool isService)
        {
            runningAsService = isService;
            autoAckEngine = new Engine();
            autoAckEngine.Load(ExecutableDirectory() + APP_AUTOACK_VMD_FILE);
            if (runningAsService)
            {
                logStream = System.IO.File.AppendText(ExecutableDirectory() + APP_LOG_FILE);
            }
        }

        static public string ExecutableDirectory()
        {
            return System.IO.Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + "\\";
        }

        public void StartHl7Server()
        {
            try
            {
                // We need to instantiate the LLP Server here since it needs to be in the thread
                // that runs the message loop.
                listener = new LlpServer();
                
                // Subscribe to the events of the LLP Server so that we will be notified and can take
                // the desired action for each type of event.
                listener.ClientConnected += new LlpServer.ClientConnectedEvent(Listener_ClientConnected);
                listener.ClientDisconnected += new LlpServer.ClientDisconnectedEvent(Listener_ClientDisconnected);
                listener.ClientErrorReceived += new LlpServer.ClientErrorReceivedEvent(Listener_ClientErrorReceived);
                listener.ErrorReceived += new LlpServer.ErrorReceivedEvent(Listener_ErrorReceived);
                listener.DataIgnored += new LlpServer.DataIgnoredEvent(Listener_DataIgnored);
                listener.MessageReceived += new LlpServer.MessageReceivedEvent(Listener_MessageReceived);

                listener.StartListen(APP_LISTEN_PORT);

                // Log information to a file.
                Log("Server started listening on " + listener.Port);

                // Run the message loop.  This is blocking.  If you wish to add functionality
                // to the start up of the service, be sure to add it *before* this call.
                iNTERFACEWARE.Chameleon.LlpServer.RunMessageLoop();
                listener.StopListen();
                listener.DisconnectAllClients();
            }
            catch (Exception ex)
            {
                Log("Error starting LLP Server" + ex.Message.ToString());
            }
        }

        public void StopHl7Server()
        {
            try
            {
                // Stop the messsage loop.  Because this is on another thread, we cannot
                // do anything else.
                listener.StopMessageLoop();
                Log("Server stopped.");
            }
            catch (Exception ex)
            {
                Log(ex.Message.ToString());
            }
        }
      
        private void Log(string output)
        {
            // TODO some type of Log Purging will be required in a real application.
            string logLine = System.DateTime.Now.ToString() + " " + output;

            if (runningAsService)
            {
                logStream.WriteLine(logLine);
                logStream.Flush();
            }
            else
            {
                System.Console.WriteLine(logLine);
            }
        }
        // This is the event we will get when a Client is connected to the Server.
        // The 'Client' object (of type LlpConnection) can be used to interact with the machine that is sending
        // the data.  There can be multiple clients connected to the server.
        private void Listener_ClientConnected(object sender, System.EventArgs e, iNTERFACEWARE.Chameleon.LlpConnection client)
        {
            Log("Client connected from: " + client.RemoteIpAddressAsString);
        }

        // This is the event we will get when a Client disconnects from the server. Note that sometimes
        // if the client is not well behaved, we may not receive this event.
        private void Listener_ClientDisconnected(object sender, System.EventArgs e, iNTERFACEWARE.Chameleon.LlpConnection client)
        {
            Log("Client disconnected from: " + client.RemoteIpAddressAsString);
        }

        // This is the event we'll get when we have an error on a given TCP/IP connection.
        private void Listener_ClientErrorReceived(object sender, System.EventArgs e, iNTERFACEWARE.Chameleon.LlpConnection client, string errorMessage, int errorCode)
        {
            Log("Client error received: " + PrettyPrintMessage(errorMessage));
        }

        // This event fires when we receive data that is not using the Lower Layer Protocol (LLP).
        // LLP is the standard protocol for sending HL7 messages and is used by Chameleon.  See the manual
        // for more details.
        private void Listener_DataIgnored(object sender, System.EventArgs e, iNTERFACEWARE.Chameleon.LlpConnection client, string data)
        {
            Log("Ignored data outside of LLP protocol received from: " + client.RemoteIpAddressAsString);
            Log(PrettyPrintMessage(data));
        }

        // This event fires if we have an error related to the listening socket itself rather
        // than a specific connection.  In practice, this event is rarely invoked.
        private void Listener_ErrorReceived(object sender, System.EventArgs e, string errorMessage, int code)
        {
            Log("Error received: " + PrettyPrintMessage(errorMessage));
        }

        // This routine converts \r --> \r\n and truncates the message if necessary so that
        // HL7 messages can be seen in the log.
        private string PrettyPrintMessage(string message)
        {
            return listener.MakeMessagePrintable(message, "\r\n", APP_MAX_MESSAGE_LOG_LENGTH);
        }

        // This event is the main entry point when we get an HL7 message from a connected 
        // client.  In this simple example, we are going to log the message to a file, and send
        // back an ACK message using auto acknowledgment.
        private void Listener_MessageReceived(object sender, System.EventArgs e, iNTERFACEWARE.Chameleon.LlpConnection client, string message)
        {
            Log("Received a message from " + client.RemoteIpAddressAsString + ":");
            Log(PrettyPrintMessage(message));
            Log("Sending ACK to " + client.RemoteIpAddressAsString);

            string ackMessage = "";
            try
            {
                // We create an acknowledgment message to send back to the client using
                // the auto-acknowledgement technique.
                ackMessage = autoAckEngine.GenerateAckFromMessage(message);                
            }
            catch (ChameleonException ex)
            {
                Log(ex.Message);
            }
            finally
            {
                if (ackMessage == "")
                {
                    Log("Generating Ack failed! Sending a hard-coded Ack.");                
                    ackMessage = "MSH|^~\\&|ACME|ACME_APP|XYZ|XYZ_APP|||NACK|XXX|P|2.4|\rMSA|AA|XXX|\r";
                }
                //We send the ACK message
                Log(PrettyPrintMessage(ackMessage));
                client.SendMessage(ackMessage);
            }
        }
    }
}
